#include "sdk.h"
#include "eloader.h"
#include "debug.h"
#include "modmgr.h"

#define MODULES_START_ADDRESS 0x08804000
#define MAX_MODULES_TO_FREE 0x20

int kill_thread(SceUID thid) 
{
    int ret = sceKernelTerminateThread(thid);
    if (ret < 0)
    {
        if (ret != (int)0x800201A2) //if thread already dormant, let's assume it's not really an error, we still want to delete it!
        {
            LOGSTR2("--> ERROR 0x%08lX TERMINATING THREAD ID 0x%08lX\n", ret, thid);
            return ret;
        }
    }

    ret = sceKernelDeleteThread(thid);
    if (ret < 0)
    {
        LOGSTR2("--> ERROR 0x%08lX DELETING THREAD ID 0x%08lX\n", ret, thid);
    }
        
    return ret;
}

int kill_event_flag(SceUID flid)
{
    int ret = sceKernelDeleteEventFlag(flid);
    if (ret < 0)
    {
        LOGSTR2("--> ERROR 0x%08lX DELETING EVENT FLAG 0x%08lX\n", ret, flid);
        return 0;
    }
    
    return 1;
} 

int kill_sema(SceUID sema)
{
    int ret = sceKernelDeleteSema(sema);
    if (ret < 0)
    {
        LOGSTR2("--> ERROR 0x%08lX DELETING SEMAPHORE 0x%08lX\n", ret, sema);
        return 0;
    }
    
    return 1;
} 

int kill_module(SceUID modid) 
{
	sceKernelStopModule(modid, 0, NULL, NULL, NULL);	
	int ret = sceKernelUnloadModule(modid);
	if (ret < 0)
	{
		LOGSTR2("--> ERROR 0x%08lX UNLOADING MODULE ID 0x%08lX\n", ret, modid);
        return 0;
	}
    return 1;
}

void DeleteAllThreads(void)
{
	u32 i;
	SceUID thids[8];
	
	thids[0] = *(SceUID*)(0x09FFEC00);
	thids[1] = *(SceUID*)(0x08B39548);
	thids[2] = *(SceUID*)(0x08BB5504);
	thids[3] = *(SceUID*)(0x08DC5CA4);
	thids[4] = *(SceUID*)(0x08DCBD54);
	thids[5] = *(SceUID*)(0x08DD1E04);
	thids[6] = *(SceUID*)(0x08DD7EB4);
	thids[7] = *(SceUID*)(0x08BB5060);
	
	/* lets kill these threads now */
	for (i = 0; i < (sizeof(thids)/sizeof(u32)); i++)
	{
		kill_thread(thids[i]);
	}
}

void DeleteAllEventFlags(void)
{
	u32 i;
	SceUID evids[2];
	
	evids[0] = *(SceUID*)(0x08BB5514);
	evids[1] = *(SceUID*)(0x08BB5550);
	
	/* killin' tiem */
	for (i = 0; i < (sizeof(evids)/sizeof(u32)); i++)
	{
		kill_event_flag(evids[i]);
	}
}

void DeleteAllSemaphores(void)
{
	u32 i;
	SceUID semaids[17];
	
	semaids[0] = *(SceUID*)(0x08E1DE30);
	semaids[1] = *(SceUID*)(0x08E26DB0);
	semaids[2] = *(SceUID*)(0x08BB504C);
	semaids[3] = *(SceUID*)(0x08BB5508);
	semaids[4] = *(SceUID*)(0x08BB550C);
	semaids[5] = *(SceUID*)(0x08BB5510);
	semaids[6] = *(SceUID*)(0x08BB554C);
	semaids[7] = *(SceUID*)(0x08BB54B4);
	semaids[8] = *(SceUID*)(0x08BB548C);
	semaids[9] = *(SceUID*)(0x08DC5CA8);
	semaids[10] = *(SceUID*)(0x08DC5CAC);
	semaids[11] = *(SceUID*)(0x08DCBD58);
	semaids[12] = *(SceUID*)(0x08DCBD5C);
	semaids[13] = *(SceUID*)(0x08DD1E08);
	semaids[14] = *(SceUID*)(0x08DD1E0C);
	semaids[15] = *(SceUID*)(0x08DD7EB8);
	semaids[16] = *(SceUID*)(0x08DD7EBC);
	
	/* lets destroy these now */
	for (i = 0; i < (sizeof(semaids)/sizeof(u32)); i++)
	{
		/* boom headshot */
		kill_sema(semaids[i]);
	}
}

void UnloadModules()
{
	// Unload user modules first
	// The more basic modules got a lower ID, so this should be the correct 
	// order to unload
	SceUID result;
	int m = PSP_MODULE_AV_G729;
	while (m >= PSP_MODULE_AV_AVCODEC)
	{
		result = sceUtilityUnloadModule(m);
		LOGSTR2("unloading utility module 0x%08lX, result 0x%08lX\n", m, result);
		m--;
	}

	m = PSP_MODULE_NET_SSL;
	while (m >= PSP_MODULE_NET_COMMON)
	{
		result = sceUtilityUnloadModule(m);
		LOGSTR2("unloading utility module 0x%08lX, result 0x%08lX\n", m, result);
		m--;
	}


	// Set inital UID to -1 and the current UID to 0
	int i;
	SceUID uids[MAX_MODULES_TO_FREE];
	uids[0] = -1;
	SceUID cur_uid = 0;
	
	/* scan through user memory looking for modules ;) */
	for (i = 0; i < (24 << 20); i += 0x400)
	{
		SceUID modid;
		
		/* check if we've got a UID */
		if ((modid = sceKernelGetModuleIdByAddress(MODULES_START_ADDRESS + i)) >= 0)
		{
			/* we do, make sure it's not just the same one */
			if (uids[cur_uid - ((cur_uid == 0) ? (0) : (1))] != modid)
			{
				/* okay add it */
				uids[cur_uid++] = modid;
			}

			if (cur_uid == MAX_MODULES_TO_FREE)
			{
				LOGSTR0("\n->WARNING: Max number of modules to unload reached\n");
				break;
			}
		}
	}
	
	/* shutdown the modules in usermode */
	for (i = cur_uid - 1; (int)i >= 0; i--)
	{
		kill_module(uids[i]);
	}
}

void FreeMem()
{
	SceUID memid;
	memid = *(SceUID*)(0x08B395C0);
	sceKernelFreePartitionMemory(memid);
	sceKernelReleaseSubIntrHandler(30,0);
}



void free_game_memory()
{
#ifdef DEBUG
    int free;
    int max_free;
	free = sceKernelTotalFreeMemSize();
    max_free = sceKernelMaxFreeMemSize();
	LOGSTR2(" FREE MEM BEFORE CLEANING: %d (max: %d)\n ", free, max_free);
#endif
	FreeMem();

	DeleteAllThreads();

	DeleteAllEventFlags();

	DeleteAllSemaphores();

    UnloadModules();

#ifdef DEBUG
	free = sceKernelTotalFreeMemSize();
    max_free = sceKernelMaxFreeMemSize();
	LOGSTR2(" FREE MEM AFTER CLEANING: %d (max: %d)\n ", free, max_free);
#endif  

	return;
}

// Those 2 functions are heavy but this avoids 2 extra syscalls that might fail
// In the future if we can have access to the "real" functions, let's remove this
SceSize sceKernelMaxFreeMemSize()
{
    SceSize size, sizeblock;
    SceUID uid;

    LOGSTR0("Call to sceKernelMaxFreeMemSize()\n");
    // Init variables
    size = 0;
    sizeblock = 1024 * 1024;

    // Check loop
    while (sizeblock)
    {
        // Increment size
        size += sizeblock;

        /* Allocate block */
        uid = sceKernelAllocPartitionMemory(2, "ValentineFreeMemMalloc", PSP_SMEM_Low, size, NULL); // Try to allocate from the lowest available address
        if (uid < 0) // Memory allocation failed
        {
            // Restore old size
            size -= sizeblock;

            // Size block / 2
            sizeblock >>= 1;
        }
        else
            sceKernelFreePartitionMemory(uid);
    }
    
    return size;
}

SceSize sceKernelTotalFreeMemSize()
{
    SceUID blocks[1024];
    u32 count,i;
    SceSize size, x;

    LOGSTR0("Call to sceKernelTotalFreeMemSize()\n");
    // Init variables
    size = 0;
    count = 0;

    // Check loop
    for (;;)
    {
        if (count >= sizeof(blocks)/sizeof(blocks[0]))
        {
            LOGSTR0("Too many blocks in sceKernelTotalFreeSize, return value will be approximate\n");
            return size;
        }

        // Find max linear size available
        x = sceKernelMaxFreeMemSize();
        if (!(x)) break;

        // Allocate ram
        blocks[count] = sceKernelAllocPartitionMemory(2, "ValentineFreeMemMalloc", PSP_SMEM_Low, x, NULL);
        if (!(blocks[count]))
        {
            LOGSTR0("Discrepency between sceKernelMaxFreeMemSize and sceKernelTotalFreeSize, return value will be approximate\n");
            return size;
        }
        
        // Update variables
        size += x;
        count++;
    }

    for (i=0;i<count;++i)
        sceKernelFreePartitionMemory(blocks[i]);

    return size;
}
